// Copyright (c) 2023 Citrix Systems, Inc. All Rights Reserved.
// You may only reproduce, distribute, perform, display, or prepare derivative works of this file pursuant to a valid license from Citrix.
// THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <fstream>
#include <cwchar>
#include <sstream>
#include <cstdlib>
#include <algorithm>
#include <linux/limits.h>

#include "Utilities.h"

std::string GetHomeDirectory()
{
    const char* home = getenv("HOME");
    if (home)
    {
        return std::string(home);
    }

    struct passwd* password_info;
    uid_t uid = getuid();
    if ((password_info = getpwuid(uid))!= NULL)
    {
        if (password_info->pw_dir)
        {
            return std::string(password_info->pw_dir);
        }
    }

    return "";
}

bool fileExists(const std::string& configFile)
{
    struct stat file_status;
    int ret_val = stat(configFile.c_str(), &file_status);
    return ret_val == 0;
}

bool isValidIcaRoot(const std::string& dir)
{
    if (!dir.length())
    {
        return false;
    }

    std::vector<std::string> expected_files;
    expected_files.push_back("AuthManagerDaemon");
    expected_files.push_back("ServiceRecord");
    expected_files.push_back("PrimaryAuthManager");
    expected_files.push_back("util/storebrowse");
    std::string file;
    for (std::vector<std::string>::iterator it = expected_files.begin() ; it != expected_files.end() ; it++)
    {
        file = dir + '/' + *it;
        if (!fileExists(file))
        {
            return false;
        }
    }

    return true;
}

std::string GetModulePath()
{
    std::string path;
    char buffer[PATH_MAX];
    ssize_t len = readlink("/proc/self/exe", buffer, sizeof(buffer) - 1);
    if (len != -1)
    {
        buffer[len] = '\0';
        path = buffer;
    }
    else
    {
        path = "<unknown>";
    }

    return path;
}

std::string GetIcaRoot()
{
    std::string icaroot_dir;
    const char* icaroot_env = getenv("ICAROOT");
    if (icaroot_env)
    {
        icaroot_dir = std::string(icaroot_env);
        if (isValidIcaRoot(icaroot_dir))
        {
            return icaroot_dir;
        }
    }

    icaroot_dir = GetModulePath();
    // This includes the executable name, we're only interested in the directory part.
    size_t found = icaroot_dir.find_last_of('/');
    if (found != std::string::npos)
    {
        icaroot_dir = icaroot_dir.substr(0, found);
    }
    if (isValidIcaRoot(icaroot_dir))
    {
        return icaroot_dir;
    }
    else
    {
        // As a utility see if this binary is running in the util directory
        // of an ICAROOT directory.
        found = icaroot_dir.find_last_of('/');
        if (found != std::string::npos)
        {
            if (icaroot_dir.substr(found + 1).compare("util") == 0)
            {
                icaroot_dir = icaroot_dir.substr(0, found);
                if (isValidIcaRoot(icaroot_dir))
                {
                    return icaroot_dir;
                }
            }
        }
    }

    // Check the default installation location.
    icaroot_dir = "/opt/Citrix/ICAClient";
    if (isValidIcaRoot(icaroot_dir))
    {
        return icaroot_dir;
    }

    // Check the old default installation location.
    icaroot_dir = "/usr/lib/ICAClient";
    if (isValidIcaRoot(icaroot_dir))
    {
        return icaroot_dir;
    }

    TRACE("No valid ICAROOT directory found");

    return "";
}

bool LoadFileIntoVector(const std::string& filePath, /* out */ std::vector<char>& outData)
{
    std::ifstream input_stream(filePath.c_str(), std::ios::binary);
    if (!input_stream)
    {
        return false;
    }
    outData.assign(std::istreambuf_iterator<char>(input_stream), std::istreambuf_iterator<char>());
    return true;
}

const char* s_Base64Chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void DataToBase64(const std::vector<char>& data, /* out */ std::string& outBase64)
{
    unsigned int index = 0;
    outBase64.clear();
    while (index < data.size())
    {
        size_t bytes_left = data.size() - index;
        unsigned char temp_base64_char;

        if (bytes_left > 2)
        {
            // Convert the next 3 bytes of data into 4 base64 chars.

            outBase64.push_back(s_Base64Chars[(data[index] >> 2) & 0x3F]);

            temp_base64_char = (data[index] & 0x03) << 4;
            index++;
            outBase64.push_back(s_Base64Chars[temp_base64_char | ((data[index] >> 4) & 0x0F)]);

            temp_base64_char = (data[index] & 0x0F) << 2;
            index++;
            outBase64.push_back(s_Base64Chars[temp_base64_char | ((data[index] >> 6) & 0x03)]);

            outBase64.push_back(s_Base64Chars[data[index] & 0x3F]);
            index++;
        }
        else
        {
            // Last 1 or 2 bytes.

            outBase64.push_back(s_Base64Chars[(data[index] >> 2) & 0x3F]);

            temp_base64_char = (data[index] & 0x03) << 4;
            index++;

            switch (bytes_left)
            {
            case 1:
                outBase64.push_back(s_Base64Chars[temp_base64_char]);
                outBase64.push_back('=');
                outBase64.push_back('=');
                break;

            case 2:
                outBase64.push_back(s_Base64Chars[temp_base64_char | ((data[index] >> 4) & 0x0F)]);
                outBase64.push_back(s_Base64Chars[(data[index] & 0x0F) << 2]);
                index++;
                outBase64.push_back('=');
                break;
            }
        }
    }
}

void ParseUrl(const std::string& urlString,
              /* out */ std::string& outProtocol,
              /* out */ std::string& outHostName,
              /* out */ unsigned int& outPort,
              /* out */ std::string& outPath,
              /* out */ std::string& outQuery)
{
    // Initialize the output parameters.
    outProtocol.clear();
    outHostName.clear();
    outPort = 0;
    outPath.clear();
    outQuery.clear();

    // Check whether the input URL is valid.
    if (urlString.empty())
    {
        return;
    }

    // Parse the URL.
    bool has_path = false;
    size_t post_domain_slash = -1;
    std::string path_and_query;

    size_t post_scheme = urlString.find("://");
    if (post_scheme != std::string::npos)
    {
        // Find next slash after "://" (at end of the domain name).
        post_domain_slash = urlString.find_first_of('/', post_scheme + 3);
        has_path = post_domain_slash != std::string::npos;
        outProtocol = urlString.substr(0, post_scheme);
        // Host name with port, path and query.
        outHostName = urlString.substr(post_scheme + 3);
    }
    else
    {
        // Host name with port, path and query.
        outHostName = urlString;
    }

    if (has_path)
    {
        // Remove path and query from host name.
        outHostName = outHostName.substr(0, (post_domain_slash - post_scheme - 3));
        path_and_query = urlString.substr(post_domain_slash);
        size_t query_begin = path_and_query.find('?');
        if (query_begin != std::string::npos)
        {
            outPath = path_and_query.substr(0, query_begin);
            outQuery = path_and_query.substr(query_begin + 1);
        }
        else
        {
            outPath = path_and_query;
        }

        if (outPath.empty())
        {
            outPath = '/';
        }
    }

    // Check for '@' in outHostName e.g. https://username:password@www.example.com/path?query#.
    size_t post_user_name = outHostName.find('@');
    if (post_user_name != std::string::npos)
    {
        outHostName = outHostName.substr(post_user_name + 1);
    }

    size_t port_ident_range = outHostName.find(':');
    if (port_ident_range != std::string::npos)
    {
        // Remove port info from host name.
        outHostName = outHostName.substr(0, port_ident_range);

        // Convert port into an integer.
        std::string port_string = outHostName.substr(port_ident_range + 1);
        size_t i = 0;
        size_t port_string_size = port_string.size();
        while (i < port_string_size)
        {
            if (isdigit(port_string.at(i)))
            {
                i++;
                continue;
            }

            break;
        }
        outPort = (i == port_string_size) ? atoi(port_string.c_str()) : -1;
    }
}

bool CrackUrl(const std::string &urlString,
              /* out */ bool& outIsHttps,
              /* out */ unsigned int& outPort,
              /* out */ std::string& outHostName,
              /* out */ std::string& outExtra)
{
    // Initialize the output parameters.
    outIsHttps = false;
    outPort = 0;
    outHostName.clear();
    outExtra.clear();

    // Parse URL.
    std::string protocol;
    std::string host_name;
    unsigned int port = 0;
    std::string path;
    std::string query;
    ParseUrl(urlString, protocol, host_name, port, path, query);

    // Check if the protocol is HTTPS.
    if (!protocol.empty())
    {
        std::transform(protocol.begin(), protocol.end(), protocol.begin(), ::tolower);
        outIsHttps = protocol == "https";
    }

    // Get port.
    if (port)
    {
        outPort = port;
    }
    else if (outIsHttps)
    {
        outPort = 443;
    }

    // Get host name.
    outHostName = host_name;

    // Get remaining part of URL.
    if (path.length() > 0)
    {
        if (!query.empty())
        {
            path = path + '?' + query;
        }

        if (urlString.at(urlString.length() - 1) == '/'
                && path.at(path.length() - 1) != '/')
        {
            path += '/';
        }

        outExtra = path;
    }
    else
    {
        outExtra = '/';
    }

    return !outHostName.empty();
}
